// This file is part of the Java Card Firewall Tester program.
// 
// Authors: 
// 
// Wojciech Mostowski, woj@cs.ru.nl
// Erik Poll, erikpoll@cs.ru.nl
// Radboud University Nijmegen
// The Netherlands
// 
// Copyright (c) Wojciech Mostowski, Erik Poll,
// Radboud University Nijmegen (RU),
// Stichting Technische Wetenschappen (STW)
// 
// The Java Card Firewall Tester has been developed for the PinPas Java
// Card project, see http://www.win.tue.nl/pinpasjc/. The program is
// distributed under the licence terms that can be found in the LICENCE
// file in the main installation directory. Please refer to the LICENCE &
// README files for further information.

package firewallserverms;

import javacard.framework.*;
import firewallinterface.*;


public class ServerApplet extends Applet
            /*JC22*/ implements MultiSelectable
{

    private static final byte RESET = (byte)0xaa;

    private ShareableObject so = null;

    public static void install(byte[] bArray, short bOffset, byte bLength) {
        // GP-compliant JavaCard applet registration
        new ServerApplet().register(bArray, (short) (bOffset + 1), bArray[bOffset]);
    }


    public ServerApplet() {
        so = new ShareableObject(TestInterface1.SIZE);
        ShareableObject.createTestObjectServerMultiselectable();
        ShareableObject.createShareableTestObjectServerMultiselectable();
    }


    public Shareable getShareableInterfaceObject(AID clientAid, byte param) {
        return so;
    }

    /*JC22*/ public void deselect(boolean appInstStillActive) { }
    /*JC22*/ public boolean select(boolean appInstAlreadyActive) {
        return true;
    }

    public void process(APDU apdu) {
        if (selectingApplet()) {
            return;
        }

        byte[] buf = apdu.getBuffer();
        apdu.setIncomingAndReceive();
        switch (buf[ISO7816.OFFSET_INS]) {
        case RESET:
            ShareableObject.resetReferences();
            return;
        default:
            ISOException.throwIt(ISO7816.SW_INS_NOT_SUPPORTED);
        }

    }

}

